import argparse
import json
from pathlib import Path

from spaghettree import Result
from spaghettree.adapters.io_wrapper import IOProtocol, IOWrapper
from spaghettree.domain.optimisation import (
    AdjMat,
    get_dwm,
    get_top_suggested_merges,
    yellow,
)
from spaghettree.domain.parsing import (
    create_call_tree,
    extract_entities_and_locations,
    filter_non_native_calls,
)
from spaghettree.domain.processing import (
    optimise_entity_positions,
)
from spaghettree.logger import logger


def main(
    src_root: str,
    *,
    new_root: str = "",
    call_tree_save_path: str = "./call_tree.json",
    optimise_src_code: bool = False,
) -> Result:
    io = IOWrapper()
    return run_process(
        io,
        src_root,
        new_root=new_root,
        optimise_src_code=optimise_src_code,
        call_tree_save_path=call_tree_save_path,
    )


def run_process(
    io: IOProtocol,
    src_root: str,
    *,
    new_root: str = "",
    call_tree_save_path: str = "./call_tree.json",
    optimise_src_code: bool = False,
) -> Result:
    logger.info(f"*** RUNNING `spaghettree` {src_root = } {new_root = } ***")
    src_code = io.read_files(src_root).unwrap()

    ent_and_locs_res = extract_entities_and_locations(src_code)
    entities, location_map = ent_and_locs_res.unwrap()
    entities_res = filter_non_native_calls(entities)
    entities = entities_res.unwrap()
    call_tree = entities_res.and_then(create_call_tree).unwrap()

    if optimise_src_code:
        res = optimise_entity_positions(
            entities=entities,
            location_map=location_map,
            call_tree=call_tree,
            src_root=src_root,
            new_root=new_root,
        ).unwrap()
    else:
        # remove any new_root so that it doesn't try to use ruff on the json
        new_root = ""
        adj_mat = AdjMat.from_call_tree(call_tree, optimise=optimise_src_code).unwrap()
        print(  # noqa: T201
            yellow(
                f"Current Directed Weighted Modularity (DWM): {get_dwm(adj_mat.mat, adj_mat.communities): .5f}"
            )
        )
        top_merges = get_top_suggested_merges(adj_mat).unwrap()

        for merge in top_merges:
            merge.display()

        res = {Path(call_tree_save_path).absolute(): json.dumps(call_tree, indent=4)}

    return io.write_files(res, ruff_root=new_root, format_bulk=optimise_src_code)


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description="Process source code from a given root, with optional relocation and optimisation."
    )

    parser.add_argument("src_root", type=str, help="Path to the source root directory.")
    parser.add_argument(
        "--new-root",
        dest="new_root",
        type=str,
        default="",
        help="Optional new root path for output (default: empty, meaning same as src_root if optimisation is enabled).",
    )
    parser.add_argument(
        "--call-tree-save-path",
        dest="call_tree_save_path",
        type=str,
        default="./call_tree.json",
        help="The location to save the generated call tree. Only used if `--optimise-src-code` isn't used. Defaults to `./call_tree.json`.",
    )
    parser.add_argument(
        "--optimise-src-code",
        dest="optimise_src_code",
        action="store_true",
        help="Enable optimisation of the source code.",
    )

    args = parser.parse_args()
    res = main(
        args.src_root,
        new_root=args.new_root,
        call_tree_save_path=args.call_tree_save_path,
        optimise_src_code=args.optimise_src_code,
    )
    if not res.is_ok():
        raise res.error



from __future__ import annotations

import os
import pathlib
from copy import deepcopy

import libcst as cst
import numpy as np
from tqdm import tqdm

from spaghettree import safe
from spaghettree.domain.entities import EntityCST
from spaghettree.domain.optimisation import AdjMat
from spaghettree.domain.visitors import EntityLocation, OnePassVisitor
from spaghettree.logger import logger


def str_to_cst(code: str) -> cst.Module:
    return cst.parse_module(code)


def cst_to_str(node: cst.CSTNode) -> str:
    return cst.Module([]).code_for_node(node)


@safe
def extract_entities_and_locations(
    src_code: dict[str, str],
) -> tuple[dict[str, EntityCST], dict[str, EntityLocation]]:
    def find_common_prefix(paths: list[str]) -> str:
        logger.debug(f"{paths = }")
        return str(pathlib.Path(os.path.commonpath(paths)).parent)

    def get_module_name(path: str, root: str) -> str:
        return os.path.splitext(path.removeprefix(root))[0].replace("/", ".").strip(".")

    common_prefix = find_common_prefix(src_code.keys())
    logger.debug(f"{common_prefix = }")

    entities: dict[str, EntityCST] = {}
    locations: dict[str, EntityLocation] = {}

    for path, data in tqdm(src_code.items(), "creating objects"):
        tree = cst.metadata.MetadataWrapper(str_to_cst(data))
        module_name = get_module_name(path, common_prefix)
        visitor = OnePassVisitor(module_name)
        tree.visit(visitor)

        entities.update(visitor.entities)
        locations.update(visitor.locations)

        import_map = {
            i.as_name: f"{i.module}.{i.as_name}" if i.module != i.as_name else i.module
            for i in visitor.imports
        }
        ent_map = {ent.name.split(".")[-1]: ent.name for ent in entities.values()}

        logger.debug(f"{import_map = }")
        logger.debug(f"{ent_map = }")

        for ent in visitor.entities.values():
            ent.resolve_calls(import_map, ent_map).add_referenced_imports(visitor.imports)

        logger.debug(f"{visitor.entities = }")
        logger.debug(f"{visitor.imports = }")

    logger.debug(f"{entities = }")
    logger.debug(f"{locations = }")
    return entities, locations


@safe
def filter_non_native_calls(
    entities: dict[str, EntityCST],
) -> dict[str, EntityCST]:
    logger.debug(f"{entities = }")
    entities = deepcopy(entities)
    return {
        name: ent.filter_native_calls(entities).resolve_native_imports()
        for name, ent in entities.items()
    }


@safe
def create_call_tree(entities: dict[str, EntityCST]) -> dict[str, list[str]]:
    logger.debug(f"{entities = }")
    return {name: ent.get_call_tree_entries() for name, ent in entities.items()}


@safe
def pair_exclusive_calls(adj_mat: AdjMat) -> AdjMat:
    adj_mat = deepcopy(adj_mat)
    matrix: np.ndarray = adj_mat.mat
    communities: list[int] = adj_mat.communities.copy()
    logger.debug(f"{adj_mat.communities = }")

    # make it so we don't weight by call count yet
    adj_bin = (matrix > 0).astype(bool)
    communities: np.ndarray = np.array(communities, dtype=int)

    changed = True
    while changed:
        changed = False

        out_deg = adj_bin.sum(axis=1)
        in_deg = adj_bin.sum(axis=0)

        rows, cols = np.where((out_deg == 1)[:, None] & adj_bin & (in_deg == 1))

        for a, b in zip(rows, cols, strict=False):
            if communities[b] != communities[a]:
                communities[communities == communities[b]] = communities[a]
                changed = True

    adj_mat.communities = communities.tolist()
    logger.debug(f"{adj_mat.communities = }")
    return adj_mat



from collections import defaultdict
from copy import deepcopy
from typing import Self

import attrs
import numpy as np

from spaghettree import safe
from spaghettree.logger import logger


@attrs.define
class AdjMat:
    mat: np.ndarray = attrs.field()
    node_map: dict[int, str] = attrs.field()
    communities: list[int] = attrs.field()
    comm_map: dict[int, str] = attrs.field(factory=dict)

    @classmethod
    @safe
    def from_call_tree(cls, call_tree: dict[str, list[str]], *, optimise: bool = True) -> Self:
        logger.debug(f"{call_tree = }")
        ent_idx, node_map, modules, mod_map = AdjMat._get_components(call_tree)
        communities = [mod_map[name] for name in modules]

        mat = AdjMat._create_adj_map(call_tree, ent_idx)

        starting_dwm = get_dwm(mat, communities)
        print(cyan("Starting DWM:"), yellow(starting_dwm))  # noqa: T201
        logger.debug(f"{starting_dwm = }")

        communities = list(node_map.keys()) if optimise else communities

        return cls(mat, node_map, communities, comm_map={v: k for k, v in mod_map.items()})

    @staticmethod
    def _create_adj_map(call_tree: dict[str, list[str]], ent_idx: dict[str, int]) -> np.ndarray:
        n = len(ent_idx)
        adj_mat = np.zeros((n, n), dtype=int)

        for caller, called in call_tree.items():
            for call in called:
                src_idx = ent_idx[caller]
                dst_idx = ent_idx[call]
                adj_mat[src_idx, dst_idx] += 1
        return adj_mat

    @staticmethod
    def _get_components(
        call_tree: dict[str, list[str]],
    ) -> tuple[dict[str, int], dict[int, str], list[str], dict[str, int]]:
        ent_idx: dict[str, int] = {node: i for i, node in enumerate(call_tree)}
        node_map: dict[int, str] = {idx: ent_name for ent_name, idx in ent_idx.items()}
        modules: list[str] = [".".join(k.split(".")[:-1]) for k in call_tree]
        unique_mods = list(dict.fromkeys(modules))
        return ent_idx, node_map, modules, {name: idx for idx, name in enumerate(unique_mods)}


@safe
def optimise_communities(adj_mat: AdjMat) -> AdjMat:
    valid_merges = get_merge_pairs(adj_mat)

    while valid_merges:
        to_merge = remove_overlapping_pairs(valid_merges)
        adj_mat.communities = apply_merges(adj_mat.communities, to_merge)
        valid_merges = get_merge_pairs(adj_mat)

    opt_dwm = get_dwm(adj_mat.mat, adj_mat.communities)

    print(cyan("Optimised DWM:"), yellow(opt_dwm))  # noqa: T201
    logger.debug(f"{opt_dwm = }")
    logger.debug(f"{adj_mat.communities = }")

    return adj_mat


@safe
def merge_single_entity_communities_if_no_gain_penalty(adj_mat: AdjMat) -> AdjMat:
    communities = np.array(adj_mat.communities)
    base_score = get_dwm(adj_mat.mat, communities)

    grouped: defaultdict[int, list[tuple[int, str]]] = defaultdict(list)

    for idx, ent_name in adj_mat.node_map.items():
        grouped[communities[idx]].append((idx, ent_name))

    updated, min_for_dir = {}, {}

    for comm, items in grouped.items():
        if len(items) == 1:
            num, name = items[0]
            dirname = ".".join(name.split(".")[:-1])
            min_for_dir[dirname] = min(num, min_for_dir.get(dirname, num))
            updated[comm] = min_for_dir[dirname]

    merge_pairs = []

    for c2, c1 in updated.items():
        merged_communities = communities.copy()
        merged_communities[merged_communities == c2] = c1
        score = get_dwm(adj_mat.mat, merged_communities)

        if gain := (score - base_score) >= 0:
            merge_pairs.append(PossibleMerge(c1, c2, gain))

    adj_mat.communities = apply_merges(adj_mat.communities, merge_pairs)
    return adj_mat


@attrs.define(eq=True, frozen=True)
class PossibleMerge:
    c1: int = attrs.field()
    c2: int = attrs.field()
    gain: float = attrs.field()


def get_merge_pairs(adj_mat: AdjMat) -> list[PossibleMerge]:
    communities = np.array(adj_mat.communities)
    unique_comms = np.unique(communities)
    base_score = get_dwm(adj_mat.mat, communities)

    merge_scores = []

    for i, c1 in enumerate(unique_comms):
        for c2 in unique_comms[i + 1 :]:
            merged_communities = communities.copy()

            merged_communities[merged_communities == c2] = c1
            score = get_dwm(adj_mat.mat, merged_communities)
            gain = score - base_score
            if gain > 0:
                merge_scores.append(PossibleMerge(c1, c2, gain))

    logger.debug(f"{merge_scores = }")
    return merge_scores


def remove_overlapping_pairs(pairs: list[PossibleMerge]) -> list[PossibleMerge]:
    pairs = sorted(pairs, key=lambda x: x.gain, reverse=True)

    selected, seen = [], set()

    for pair in pairs:
        if pair.c1 not in seen and pair.c2 not in seen:
            selected.append(pair)
            seen.add(pair.c1)
            seen.add(pair.c2)

    logger.debug(f"{selected = }")
    return selected


def apply_merges(communities: list[int], pairs: list[PossibleMerge]) -> list[int]:
    communities = np.array(communities)

    for pair in pairs:
        communities[communities == pair.c2] = pair.c1
    return communities.tolist()


def get_dwm(mat: np.ndarray, communities: list[int]) -> float:
    out_degree = mat.sum(axis=0)
    in_degree = mat.sum(axis=1)
    total_edges = out_degree.sum()

    if total_edges == 0:
        return 0

    communities = np.asarray(communities)
    community_mat = np.equal.outer(communities, communities)

    expected_matrix = np.outer(out_degree, in_degree) / total_edges
    modularity_matrix = (mat - expected_matrix)[community_mat]
    return modularity_matrix.sum() / total_edges


@attrs.define(eq=True, frozen=True)
class SuggestedMerge:
    entity: str = attrs.field()
    target_community: str = attrs.field()
    gain: float = attrs.field(converter=float)

    def display(self) -> None:
        print(  # noqa: T201
            cyan(f"{self.entity}"),
            yellow("->"),
            cyan(f"{self.target_community}"),
            green(f"+{self.gain:.3f}"),
        )


@safe
def get_top_suggested_merges(adj_mat: AdjMat, top_n: int = 5) -> list[SuggestedMerge]:
    adj_mat = deepcopy(adj_mat)
    matrix: np.ndarray = adj_mat.mat
    communities: list[int] = adj_mat.communities.copy()
    comm_choices: set[int] = set(communities)

    starting_score = get_dwm(matrix, communities)

    suggested_merges = []

    for idx, comm in enumerate(communities):
        best_choice, best_score = None, starting_score

        for choice in comm_choices:
            if comm == choice:
                continue
            new_comms = communities.copy()
            new_comms[idx] = choice
            score = get_dwm(matrix, new_comms)

            if score > best_score:
                best_choice, best_score = choice, score

        if best_choice is not None:
            suggested_merges.append(
                SuggestedMerge(
                    adj_mat.node_map.get(idx),
                    adj_mat.comm_map.get(best_choice),
                    best_score - starting_score,
                )
            )

    return sorted(suggested_merges, key=lambda x: -x.gain)[:top_n]


def yellow(inp_str: str) -> str:
    return f"\033[33m{inp_str}\033[0m"


def cyan(inp_str: str) -> str:
    return f"\033[36m{inp_str}\033[0m"


def green(inp_str: str) -> str:
    return f"\033[32m{inp_str}\033[0m"
